/*
   fdnpkg - FreeDOS Network Package manager
   Copyright (C) Mateusz Viste 2012
   
   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program. If not, see <http://www.gnu.org/licenses/>.
*/


#include <ctype.h>    /* tolower() */
#include <stdio.h>    /* printf() */
#include <stdlib.h>   /* getenv() */
#include <string.h>   /* strcasecmp() */
#include "htget.h"    /* htget() */
#include "kittendj.h" /* required for kitten subsystem */
#include "kprintf.h"  /* kitten_printf() */
#include "loadconf.h" /* loadconf() */
#include "pkgdb.h"    /* createdb(), findpkg()... */
#include "pkgsrch.h"  /* pkgsrch() */
#include "pkginst.h"  /* pkginstall() */
#include "pkgrem.h"   /* pkgrem() */
/* #include "srchpath.h"  */ /* searchpath() */

#define PVER "0.9"
#define PDATE "2012"


extern char *wattcpVersion(); /* provided by wattcp to poll its version */

void printhelp() {
  puts("FDNPKG v" PVER " Copyright (C) Mateusz Viste " PDATE);
  puts("This is a network package manager for FreeDOS.");
  puts("");
  puts("Usage: FDNPKG action [parameters]");
  puts("");
  puts("Where action is one of the following:");
  puts(" search [string]   - Search net repositories for package containing 'string'");
  puts(" vsearch [string]  - Same as 'search', but prints also source repositories");
  puts(" install pkgname   - Install the package 'pkgname' (or local zip file)");
  puts(" remove pkgname    - Remove the package 'pkgname'");
  puts(" listrep           - print out the list of configured repositories");
  puts(" license           - print out the license of this program");
  puts("");
  puts("FDNPKG is linked against WatTCP version below:");
  puts(wattcpVersion());
}


void printlic() {
  puts("FDNPKG - FreeDOS Network Package manager");
  puts("Copyright (C) Mateusz Viste 2012");
  puts("");
  puts("This program is free software: you can redistribute it and/or modify it under");
  puts("the terms of the GNU General Public License as published by the Free Software");
  puts("Foundation, either version 3 of the License, or (at your option) any later");
  puts("version.");
  puts("");
  puts("This program is distributed in the hope that it will be useful, but WITHOUT ANY");
  puts("WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A");
  puts("PARTICULAR PURPOSE. See the GNU General Public License for more details.");
  puts("");
  puts("You should have received a copy of the GNU General Public License along with");
  puts("this program. If not, see <http://www.gnu.org/licenses/>.");
  puts("");
}


enum actions {
  ACTION_HELP,
  ACTION_SEARCH,
  ACTION_INSTALL,
  ACTION_INSTALL_LOCALFILE,
  ACTION_REMOVE,
  ACTION_LISTREP,
  ACTION_LICENSE
};


#define QUIT(x) \
  kittenclose(); \
  return(x);


int main(int argc, char **argv) {
  int htgetres;
  char *tempdir;
  char cfgfile[256];
  char *dosdir;
  #define MAXREPS 64
  char *repolist[MAXREPS];
  int repolistcount;
  int verbosemode = 0;
  struct pkgdb *pkgdb = NULL;
  enum actions action = ACTION_HELP;

  /* First init KITTEN */
  kittenopen("FDNPKG");

  /* check if %TEMP% is set, and retrieve it */
  tempdir = getenv("TEMP");
  if (tempdir == NULL) {
    printf("%%TEMP%% not set! You should make it point to a writeable directory.\r\n");
    printf("Example: SET TEMP=C:\\TEMP\r\n");
    QUIT(0)
  }

  /* check if %DOSDIR% is set, and retrieve it */
  dosdir = getenv("DOSDIR");
  if (dosdir == NULL) {
    puts("%%DOSDIR%% not set! You should make it point to the FreeDOS main directory.");
    puts("Example: SET DOSDIR=C:\\FDOS");
    QUIT(0)
  }

  /* check if UNZIP is present in the path - see FDPKG source for hints (line 479 of func.c) */
  /* if (searchpath("unzip.exe", "PATH") == NULL) {
    printf("Could not find or run UNZIP.EXE from within the PATH.\r\n");
    printf("Make sure UNZIP.EXE is properly installed on your system.\r\n");
    QUIT(0)
  } */

  sprintf(cfgfile, "%s\\bin\\fdnpkg.cfg", dosdir);

  /* parse cli parameters */
  if (argc > 1) { /* fdnpkg action [param] */
    if (strcasecmp(argv[1], "search") == 0) {
        if (argc > 3) {
            printf("Invalid number of arguments. Run fdnpkg without any parameter for help.\r\n");
            QUIT(0)
          } else {
            action = ACTION_SEARCH;
        }
      } else if (strcasecmp(argv[1], "vsearch") == 0) {
        if (argc > 3) {
            printf("Invalid number of arguments. Run fdnpkg without any parameter for help.\r\n");
            QUIT(0)
          } else {
            action = ACTION_SEARCH;
            verbosemode = 1;
        }
      } else if (strcasecmp(argv[1], "install") == 0) {
        if (argc != 3) {
            printf("Invalid number of arguments. Run fdnpkg without any parameter for help.\r\n");
            QUIT(0)
          } else {
            int arglen = strlen(argv[2]);
            action = ACTION_INSTALL;
            if (arglen > 4) {
              if ((argv[2][arglen - 4] == '.') && (tolower(argv[2][arglen - 3]) == 'z') && (tolower(argv[2][arglen - 2]) == 'i') && (tolower(argv[2][arglen - 1]) == 'p')) { /* if argument ends with '.zip', then it's a local package file */
                action = ACTION_INSTALL_LOCALFILE;
              }
            }
        }
      } else if (strcasecmp(argv[1], "remove") == 0) {
        if (argc != 3) {
            printf("Invalid number of arguments. Run fdnpkg without any parameter for help.\r\n");
          } else {
            pkgrem(argv[2], dosdir);
        }
        QUIT(0);
      } else if (strcasecmp(argv[1], "listrep") == 0) {
        if (argc != 2) {
            printf("Invalid number of arguments. Run fdnpkg without any parameter for help.\r\n");
            QUIT(0)
          } else {
            action = ACTION_LISTREP;
        }
      } else if (strcasecmp(argv[1], "license") == 0) {
        if (argc != 2) {
            puts("Invalid number of arguments. Run fdnpkg without any parameter for help.");
          } else {
            printlic();
        }
        QUIT(0)
    }
  }

  if (action == ACTION_HELP) {
    printhelp();
    QUIT(0)
  }

  /* if we install from a local file, do it and quit */
  if (action == ACTION_INSTALL_LOCALFILE) {
    char pkgname[64];
    int t, lastpathdelim = 0, u = 0;
    for (t = 0; argv[2][t] != 0; t++) {
      if ((argv[2][t] == '/') || (argv[2][t] == '\\')) lastpathdelim = t;
    }
    /* copy the filename into pkgname (without path elements) */
    for (t = lastpathdelim + 1; argv[2][t] != 0; t++) pkgname[u++] = argv[2][t];
    pkgname[u] = 0; /* terminate the string */
    /* truncate the file's extension (.zip) */
    for (t = u; t > 0; t--) {
      if (pkgname[t] == '.') {
        pkgname[t] = 0;
        break;
      }
    }
    pkginstall(NULL, pkgname, dosdir, tempdir, argv[2]);
    QUIT(0)
  }

  /* Load the list of package repositories */
  repolistcount = loadconf(cfgfile, repolist, MAXREPS);
  if (repolistcount < 0) return(5);
  if (repolistcount == 0) {
    printf("No repository is configured. You need at least one.\r\n");
    printf("You should place in your configuration file at least one entry of such form:\r\n");
    printf("REPO www.freedos.org/repo\r\n");
    printf("\r\n");
    QUIT(0)
  }

  if (action == ACTION_LISTREP) { /* if all we wanted was to print out repositories... */
      int x;
      puts("The list of configured fdnpkg repositories follows:");
      for (x = 0; x < repolistcount; x++) puts(repolist[x]);
      puts("");
    } else { /* other actions: search, install */
      pkgdb = createdb();
      if (pkgdb != NULL) {
        int x;
        char tempfile[512];
        char repoindex[512];
        sprintf(tempfile, "%s\\fdnpkg.tmp", tempdir);
        /* download every repo index into %TEMP% and load them in RAM */
        for (x = 0; x < repolistcount; x++) {
          printf("Refreshing %s...\r\n", repolist[x]);
          sprintf(repoindex, "%sindex.lst", repolist[x]);
          htgetres = htget(repoindex, NULL, NULL, 0, tempfile);
          if (htgetres != 0) {
              puts("Repository download failed!");
            } else {
              if (loaddb(pkgdb, "c:\\temp\\fdnpkg.tmp", repolist[x]) != 0) puts("An error occured while trying to load repository from tmp file...");
          }
        }
        if (action == ACTION_SEARCH) { /* for search: iterate through the sorted db, and print out all packages that match the pattern */
            if (argc == 3) { /* a search term has been provided */
                pkgsearch(pkgdb, argv[2], verbosemode);
              } else {
                pkgsearch(pkgdb, NULL, verbosemode);
            }
          } else if (action == ACTION_INSTALL) { /* for install: iterate through the sorted db until you find package pkgname. Then, install it or update it. */
            pkginstall(pkgdb, argv[2], dosdir, tempdir, NULL);
        }
        /* free memory of the pkg database */
        freedb(&pkgdb);
      } /* pkgdb != NULL */
  } /* action == ACTION_LISTREP */

  freeconf(repolist, repolistcount);
  QUIT(0);
}
